/*
 * Driver IDs
 * Copyright (C) 2012 Vasily Khoruzhick <anarsoul@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef __DRIVER_IDS
#define __DRIVER_IDS

enum {
	UPEKTS_ID	= 1,
	URU4000_ID	= 2,
	AES4000_ID	= 3,
	AES2501_ID	= 4,
	UPEKTC_ID	= 5,
	AES1610_ID	= 6,
	FDU2000_ID	= 7,
	VCOM5S_ID	= 8,
	UPEKSONLY_ID	= 9,
	VFS101_ID	= 10,
	VFS301_ID	= 11,
	AES2550_ID	= 12,
	UPEKE2_ID	= 13,
	AES1660_ID	= 14,
	AES2660_ID	= 15,
	AES3500_ID	= 16,
	UPEKTC_IMG_ID	= 17,
	ETES603_ID	= 18,
	VFS5011_ID	= 19,
};

#endif
